from .pooltype import PoolType
from abc import abstractmethod
from .delegate import Delegate
from .exceptions import PoolException
from ...kernel.agent.Agent import Agent
from .responseaction import ResponseAction
from .notifyfreeaction import NotifyFreeAction

class PoolController(Agent):

    __type = None
    __bufferSize = None
    
    def __init__(self, agentID, type, bufferSize):
        self.__type = type
        self.__bufferSize = bufferSize
        super().__init__(agentID)

    def setUp(self):
        self.state = {
            'social': True,
            'freeList': [],
            'requestDict': {},
            'type': self.__type,
            'bufferSize': self.__bufferSize
        }
        self.addBehavior('Delegate')
        if self.__type == PoolType.BLOCK:
            self.bindAction('Delegate', 'delegate', Delegate())
        self.addBehavior('Notify')
        self.bindAction('Notify', 'notify', NotifyFreeAction())
        self.addBehavior('Response')
        self.bindAction('Response', 'response', ResponseAction())
        self.build()

    def bindDelegateAction(self, action):
        if self.__type == PoolType.NO_BLOCK:
            self.bindAction('Delegate', 'delegate', action)
        else:
            raise PoolException('[Warn, bindDelegateAction]: The controller is a blocking type. No need to define delegator')

    def suscribeAgent(self, agent):
        agent.state['controller'] = self.id
        agent.state['controllerType'] = 'POOL'
        self.state['freeList'].append(agent.id)
        actions = agent.getActions()
        for action in actions:
            action.isPool = True
            action.enableResponse = (self.__type == PoolType.BLOCK)
    
    @abstractmethod
    def build(self):
        pass



from .task import Task
from abc import abstractmethod
from .exceptions import WorkerException
from ...kernel.agent.Agent import Agent
from .timeoutaction import TimeoutAction

class Worker(Agent):

    __taskList = []

    def setUp(self):
        self.state = {
            'controller': None,
            'controllerType': None
        }
        self.addBehavior('Task')
        self.addBehavior('Timeout')
        self.bindAction('Timeout', 'timeout', TimeoutAction())
        self.build()

    def bindTask(self, action):
        if isinstance(action, Task):
            self.__taskList.append(action)    
            self.bindAction('Task', 'task', action)
        else:
            raise WorkerException('[Warn, bindTask]: The action must inherit from the task type')

    def getActions(self):
        return self.__taskList

    @abstractmethod
    def build(self):
        pass



# -*- coding: utf-8 -*-
"""
----------------------------------------------------------
------------------------- PBESA --------------------------
----------------------------------------------------------

@autor AKEN & SIDRE
@version 3.0.1
@date 27/07/20
"""

# --------------------------------------------------------
# Define resources
# --------------------------------------------------------
import time
import random
import traceback
from threading import Thread
from ...kernel.agent.exceptions import ActionException

# --------------------------------------------------------
# Define component
# --------------------------------------------------------
class BehaviorExe(Thread):
    """ Behavior executor component """

    let = None
    alive = None
    queue = None

    def __init__(self, queue):
        self.let = False
        self.alive = True
        self.queue = queue
        super().__init__()

    def run(self):
        while self.alive:
            evt = self.queue.get()
            if self.let:
                self.queue.task_done()
                try:
                    evt['action'].execute(evt['data'])
                except Exception as e:
                    evt['action'].catchException(e)
            else:
                time.sleep(1)                

    def setLet(self, val):
        self.let = val

    def setAlive(self, val):
        self.alive = val

    def finalize(self):
        self.let = False
        self.queue.put(None)



from abc import abstractmethod
from ...kernel.agent.Agent import Agent
from .timeoutaction import TimeoutAction
from .delegateaction import DelegateAction
from .responseaction import ResponseAction

class LinealController(Agent):
    
    def setUp(self):
        self.state = {
            'social': True,
            'freeList': [],
            'agentList': [],
            'checkDict': {},
            'gateway': None,
            'timeout': False
        }
        self.addBehavior('Timeout')
        self.bindAction('Timeout', 'timeout', TimeoutAction())
        self.addBehavior('Delegate')
        self.addBehavior('Response')
        self.build()

    def bindDelegateAction(self, action):
        self.bindAction('Delegate', 'delegate', action)

    def bindResponseAction(self, action):
        self.bindAction('Response', 'response', action)

    def suscribeAgent(self, agent):
        agent.state['controller'] = self.id
        agent.state['controllerType'] = 'LINEAL'
        self.state['checkDict'][agent.id] = None
        self.state['freeList'].append(agent.id)
        self.state['agentList'].append(agent.id)
        actions = agent.getActions()
        for action in actions:
            action.isPool = False
            action.enableResponse = True
    
    def reset(self):
        self.state['freeList'] = self.state['agentList'].clone()
        for ag in self.state['freeList']:
            self.state['checkDict'][ag] = None

    @abstractmethod
    def build(self):
        pass



from abc import abstractmethod
from ...kernel.agent.Action import Action

# --------------------------------------------------------
# Define Action
# --------------------------------------------------------
class Delegate(Action):
    """ An action is a response to the occurrence of an event """

    def execute(self, data):
        """ 
        Response.
        @param data Event data 
        """
        ag = self.agent.state['freeList'].pop(0)
        self.agent.state['requestDict'][ag] = {
            'gateway': data['gateway'],
            'dtoList': []
        }
        self.adm.sendEvent(ag, 'task', data['dto'])

    def catchException(self, exception):
        """
        Catch the exception.
        @param exception Response exception
        """
        pass
